import 'package:flutter_test/flutter_test.dart' as flutter_test show test;
import 'package:flutter_test/flutter_test.dart' hide test;
import 'package:meta/meta.dart';

import '../theatre.dart';

extension FlameFinds on CommonFinders {
  Finder byGame<T extends Stage>() {
    return find.byWidgetPredicate((widget) => widget is TheatreWidget<T>);
  }
}

typedef GameCreateFunction<T extends Stage> = T Function();
typedef VerifyFunction<T extends Stage> = dynamic Function(T);

typedef GameWidgetCreateFunction<T extends Stage> = TheatreWidget<T> Function(
  T game,
);
typedef WidgetVerifyFunction<T extends Stage> = Future<void> Function(
  T,
  WidgetTester,
);
typedef WidgetSetupFunction<T extends Stage> = Future<void> Function(
  T,
  WidgetTester,
);
typedef PumpWidgetFunction<T extends Stage> = Future<void> Function(
  TheatreWidget<T>,
  WidgetTester tester,
);

class StageTester<T extends Stage> {
  final GameCreateFunction<T> createStage;

  final GameWidgetCreateFunction<T>? createTheatreWidget;

  final PumpWidgetFunction<T>? pumpWidget;

  /// Override the game size to be provided during `onGameResize`.
  /// By default it will be a 500x500 square.
  final Vector2? canvasSize;

  StageTester(
    this.createStage, {
    this.canvasSize,
    this.createTheatreWidget,
    this.pumpWidget,
  });

  T initializeGame() {
    final stage = createStage();
    final size = canvasSize ?? Vector2.all(500);

    stage.onCanvasResize(size);
    stage.onMount();
    stage.update();

    return stage;
  }

  @isTest
  void test(String description, VerifyFunction<T> verify, {String? skip}) {
    flutter_test.test(
      description,
      () => verify(initializeGame()),
      skip: skip,
    );
  }

  @isTest
  void testGameWidget(
    String description, {
    WidgetSetupFunction<T>? setUp,
    WidgetVerifyFunction<T>? verify,
  }) {
    testWidgets(description, (tester) async {
      final stage = createStage();

      await tester.runAsync(() async {
        final gameWidget = createTheatreWidget?.call(stage) ?? TheatreWidget(stage: stage);

        final _pump = pumpWidget ??
            (TheatreWidget<T> _gameWidget, WidgetTester _tester) => _tester.pumpWidget(_gameWidget);

        await _pump(gameWidget, tester);
        await tester.pump();

        if (setUp != null) {
          await setUp.call(stage, tester);
          await tester.pump();
        }
      });

      if (verify != null) {
        await verify(stage, tester);
      }
    });
  }

  StageTester<T> configure({
    GameCreateFunction<T>? createStage,
    Vector2? canvasSize,
    GameWidgetCreateFunction<T>? createTheatreWidget,
    PumpWidgetFunction<T>? pumpWidget,
  }) {
    return StageTester<T>(
      createStage ?? this.createStage,
      canvasSize: canvasSize ?? this.canvasSize,
      createTheatreWidget: createTheatreWidget ?? this.createTheatreWidget,
      pumpWidget: pumpWidget ?? this.pumpWidget,
    );
  }
}
